#    Interleave Playlist
#    Copyright (C) 2021-2022 Thomas Sweeney
#    This file is part of Interleave Playlist.
#    Interleave Playlist is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    Interleave Playlist is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import os
import re
import typing
from collections.abc import Iterable
from datetime import datetime
from pathlib import Path
from typing import Any

from crontab import CronTab
from ruamel.yaml import YAML, YAMLError

from interleave_playlist.core import PlaylistEntry
from interleave_playlist.persistence import Location, Group, Timed, _STATE_FILE


class InvalidInputFile(Exception):
    pass


class LocationNotFound(Exception):
    pass


def get_locations() -> list[Location]:
    input_: dict = _get_input(get_last_input_file())
    locations = []
    for loc in input_['locations']:
        if 'disabled' in loc and loc['disabled'] is True:
            continue
        options = [loc, input_]
        locations.append(
            Location(
                loc['name'],
                Group(
                    loc['name'],
                    loc['name'],
                    _nested_get('priority', options),
                    _nested_get('whitelist', options),
                    _nested_get('blacklist', options),
                    _get_timed(loc['timed']) if 'timed' in loc else None,
                ),
                loc['additional'] if 'additional' in loc else [],
                _nested_get('regex', options),
                _get_group_list(loc['groups'], options) if 'groups' in loc else []
            )
        )
    return locations


def get_watched_file_name() -> str:
    fn = get_last_input_file() + '.watched.txt'
    if not os.path.exists(fn):
        with open(fn, 'w'):
            pass
    return fn


def drop_groups(entries: Iterable[PlaylistEntry]) -> None:
    input_ = _get_input(get_last_input_file())
    for entry in entries:
        location = next(filter(lambda i: i['name'] == entry.location.name, input_['locations']))
        if entry.group.name == entry.location.name:
            location['disabled'] = True
            continue
        group_name = entry.group.name
        blacklist: list[str] = location.setdefault('blacklist', [])
        if group_name not in blacklist:
            blacklist.append(group_name)
    yaml = YAML()
    yaml.dump(input_, Path(get_last_input_file()))


def _get_group_list(data: list[dict[str, Any]], additional_options: list[dict[str, Any]]) \
        -> list[Group]:
    groups = []
    for g in data:
        options = [g, *additional_options]
        groups.append(Group(
            g['name'],
            _nested_get('name', additional_options),
            _nested_get('priority', options),
            _nested_get('whitelist', options),
            _nested_get('blacklist', options),
            _nested_get('timed', options),
        ))
    return groups


def _get_timed(d: dict[str, Any]) -> Timed:
    return Timed(
        datetime.fromisoformat(d['start']),
        CronTab(d['cron']),
        d.get('first'),
        d.get('amount'),
        d.get('start-at-cron'),
    )


def _get_input(input_file: str) -> dict[str, Any]:
    try:
        with open(input_file, 'r') as f:
            yaml = YAML()
            yaml.preserve_quotes = True  # type: ignore
            yml = yaml.load(f)
        _validate_group(yml)
        if 'locations' not in yml:
            raise InvalidInputFile('Input requires "locations"')
        if not isinstance(yml['locations'], list):
            raise InvalidInputFile('"locations" must be a list')
        for loc in yml['locations']:
            if 'name' not in loc:
                raise InvalidInputFile('Locations must have names')
            if not isinstance(loc['name'], str):
                raise InvalidInputFile('Location names must be strings')
            if not os.path.exists(loc['name']):
                raise LocationNotFound(loc['name'])
            if 'disabled' in loc and not isinstance(loc['disabled'], bool):
                raise InvalidInputFile('"disabled" must be a boolean')
            _validate_group(loc)
            _validate_timed(loc)
            if 'groups' in loc:
                if not isinstance(loc['groups'], list):
                    raise InvalidInputFile('groups must be a list')
                _validate_groups(loc['groups'])
            if 'additional' in loc:
                if not isinstance(loc['additional'], list):
                    raise InvalidInputFile('"additional" must be a list')
                for i in loc['additional']:
                    if not isinstance(i, str):
                        raise InvalidInputFile('"additional" entries must be strings')

    except YAMLError as e:
        raise InvalidInputFile("Unable to parse input file") from e
    return typing.cast(dict[str, Any], yml)


def _validate_wb_list(d: dict, wb_list: str) -> None:
    if wb_list in d:
        if not isinstance(d[wb_list], list):
            raise InvalidInputFile(wb_list + ' must be a list')
        for white in d[wb_list]:
            if not isinstance(white, str):
                raise InvalidInputFile(wb_list + ' entries must be strings')


def _validate_whitelist(d: dict) -> None:
    _validate_wb_list(d, 'blacklist')


def _validate_blacklist(d: dict) -> None:
    _validate_wb_list(d, 'whitelist')


def _validate_regex(d: dict) -> None:
    if d.get('regex'):
        if not isinstance(d['regex'], str):
            raise InvalidInputFile("Regex must be a string")
        try:
            re.compile(d['regex'])
        except re.error:
            raise InvalidInputFile("Regex is invalid")


def _validate_timed(d: dict) -> None:
    if d.get('timed'):
        if 'start' not in d['timed']:
            raise InvalidInputFile('Usage of timed requires a start datetime')
        if 'cron' not in d['timed']:
            raise InvalidInputFile('Usage of timed requires a cron rule')
        try:
            datetime.fromisoformat(d['timed']['start'])
        except (TypeError, ValueError) as e:
            raise InvalidInputFile('Invalid start date format. Requires ISO string') from e
        try:
            CronTab(d['timed']['cron'])
        except ValueError as e:
            raise InvalidInputFile('Invalid cron format') from e
        if 'first' in d['timed'] and not isinstance(d['timed']['first'], int):
            raise InvalidInputFile('timed.first must be an integer')
        if 'amount' in d['timed'] and not isinstance(d['timed']['amount'], int):
            raise InvalidInputFile('timed.amount must be an integer')


def _validate_priority(d: dict) -> None:
    if 'priority' in d and not isinstance(d['priority'], int):
        raise InvalidInputFile('priority must be an integer')


def _validate_group(group: dict) -> None:
    validators = [
        _validate_whitelist,
        _validate_blacklist,
        _validate_regex,
        _validate_priority,
    ]
    for validator in validators:
        validator(group)


def _validate_groups(groups: list[dict]) -> None:
    for group in groups:
        _validate_group(group)


def _nested_get(key: str, options: list[dict[str, Any]]) -> Any:
    for option in options:
        if key in option:
            return option.get(key)


def get_last_input_file() -> str:
    return typing.cast(str, _get_state()['last-input-file'])


def set_last_input_file(input_file: str) -> None:
    _get_input(input_file)  # ensure that the input can be read
    _set_state('last-input-file', input_file)


def _get_state() -> dict[str, Any]:
    with open(_STATE_FILE, 'r') as f:
        return typing.cast(dict[str, Any], json.load(f))


def _set_state(key: str, val: Any) -> None:
    state = _get_state()
    state[key] = val
    with open(_STATE_FILE, 'w') as f:
        json.dump(state, f)



#    Interleave Playlist
#    Copyright (C) 2021-2022 Thomas Sweeney
#    This file is part of Interleave Playlist.
#    Interleave Playlist is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    Interleave Playlist is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from typing import TypeVar

T = TypeVar('T')
_MARGIN = 10e-6


def interleave(a: list[T], b: list[T]) -> list[T]:
    smaller, larger = sorted([a, b], key=lambda ab: len(ab))
    if not smaller:
        return larger
    group_count = len(smaller) + 1
    group_size = len(larger) // group_count + 1
    surplus = len(larger) - group_count * (group_size - 1)
    surplus_per_group = surplus / group_count
    result = []
    larger_idx = 0

    def _append_larger() -> None:
        nonlocal larger_idx
        use_surplus = surplus_per_group and (group_idx+1+_MARGIN) % (1/surplus_per_group) < 1
        surplus_offset = 1 if use_surplus else 0
        for i in range(group_size - 1 + surplus_offset):
            result.append(larger[larger_idx])
            larger_idx += 1

    group_idx = 0
    for smaller_idx in range(len(smaller)):
        _append_larger()
        result.append(smaller[smaller_idx])
        group_idx += 1
    _append_larger()
    return result


# Sort by minimum group size difference
def interleave_all(groups: list[list[T]]) -> list[T]:
    groups = [group for group in groups if group]  # just make debugging easier
    sorted_groups = sorted(groups, key=lambda l: len(l))
    while len(sorted_groups) > 1:
        min_diff: int = sys.maxsize
        min_i: int = -1
        for i, (left, right) in enumerate(zip(sorted_groups, sorted_groups[1:])):
            diff: int = abs(len(left) - len(right))
            if min_diff > diff:
                min_diff = diff
                min_i = i
        a: list[T] = sorted_groups.pop(min_i)  # smaller or equal
        b: list[T] = sorted_groups.pop(min_i)
        i = 0
        while i < len(sorted_groups):  # interleave smaller groups to reduce size diff
            if abs(len(sorted_groups[i]) + len(a) - len(b)) <= abs(len(a) - len(b)):
                a = interleave(a, sorted_groups.pop(i))
            else:
                i += 1
        interleaved = interleave(a, b)
        len_before_insert = len(sorted_groups)
        for i, elem in enumerate(sorted_groups):
            if len(elem) > len(interleaved):
                sorted_groups.insert(i, interleaved)
                break
        if len_before_insert == len(sorted_groups):
            sorted_groups.append(interleaved)
    return sorted_groups[0] if len(sorted_groups) > 0 else []



#    Interleave Playlist
#    Copyright (C) 2021-2022 Thomas Sweeney
#    This file is part of Interleave Playlist.
#    Interleave Playlist is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    Interleave Playlist is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import csv
import os
from os import path

from interleave_playlist.core.playlist import FileGroup, get_playlist, PlaylistEntry
from interleave_playlist.persistence import input_
from interleave_playlist.persistence import settings


def get_watched() -> list[FileGroup]:
    with open(input_.get_watched_file_name(), 'r') as f:
        rows = csv.reader(f)
        watched_list: list[FileGroup] = []
        for row in rows:
            if len(row) == 0:
                continue
            item = row[0]
            if len(row) > 1:
                group = row[1]
            else:
                group = ''
            watched_list.append((item, group))
        return watched_list


def add_watched(add: list[PlaylistEntry]) -> None:
    new_watched_list: list[FileGroup] = _clean_watched_list([])
    for a in add:
        new_watched_list.append((path.basename(a.filename), a.group.name))
    _write_new_watched_list(new_watched_list)


def remove_watched(remove: list[PlaylistEntry]) -> None:
    remove_names = [path.basename(i.filename) for i in remove]
    new_watched_list: list[FileGroup] = _clean_watched_list(remove_names)
    _write_new_watched_list(new_watched_list)


def _get_temp_file_name() -> str:
    return input_.get_watched_file_name() + '.tmp'


def _get_basename_playlist() -> list[PlaylistEntry]:
    return [i for i in get_playlist(input_.get_locations(), [], use_cache=True)]


def _clean_watched_list(remove_names: list[str]) -> list[FileGroup]:
    full_playlist = _get_basename_playlist()
    watched_list = get_watched()
    new_watched_list: list[FileGroup] = []
    max_watched_remembered = settings.get_max_watched_remembered()
    watched_remembered = 0
    for row in reversed(watched_list):
        new_watched_list_len = len(new_watched_list)
        if row[0] not in remove_names:
            for item in full_playlist:
                if row[0].strip() == os.path.basename(item.filename).strip():
                    new_watched_list.append(row)
                    break
            if (new_watched_list_len == len(new_watched_list)
                    and watched_remembered < max_watched_remembered):
                watched_remembered += 1
                new_watched_list.append(row)
    new_watched_list.reverse()
    return new_watched_list


def _write_new_watched_list(new_watched_list: list[FileGroup]) -> None:
    with open(_get_temp_file_name(), 'w') as tmp:
        writer = csv.writer(tmp, quoting=csv.QUOTE_ALL)
        writer.writerows(new_watched_list)
    os.replace(_get_temp_file_name(), input_.get_watched_file_name())



#    Interleave Playlist
#    Copyright (C) 2021-2022 Thomas Sweeney
#    This file is part of Interleave Playlist.
#    Interleave Playlist is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    Interleave Playlist is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import platform
import subprocess
import sys
from math import log10, ceil

from PySide6.QtWidgets import QMessageBox

from interleave_playlist.core.playlist import get_playlist, PlaylistEntry
from interleave_playlist.persistence import input_
from interleave_playlist.persistence.watched import get_watched


# https://stackoverflow.com/questions/434597/open-document-with-default-os-application-in-python-both-in-windows-and-mac-os
def open_with_default_application(filepath: str) -> None:
    if platform.system() == 'Darwin':       # macOS
        subprocess.call(('open', filepath))
    # This is actually ridiculous https://github.com/python/mypy/issues/8547
    elif sys.platform == 'win32' or sys.platform == 'cygwin':  # Windows
        os.startfile(filepath)
    else:                                   # linux variants
        subprocess.call(('xdg-open', filepath))


def _create_playlist(search_filter: str = "", use_cache: bool = False) -> list[PlaylistEntry]:
    def show_warning(text: str) -> None:
        msg_box = QMessageBox()
        msg_box.setWindowTitle('Error')
        msg_box.setText(text)
        msg_box.setIcon(QMessageBox.Warning)
        msg_box.show()
    try:
        return get_playlist(input_.get_locations(), get_watched(), search_filter, use_cache)
    except FileNotFoundError:
        show_warning(f'Input yml file not found: {input_.get_last_input_file()}\n\n'
                     'Please create or find file and open it')
    except input_.InvalidInputFile as e:
        show_warning(f'Error reading yml file. Please fix it and try again\n'
                     f'{input_.get_last_input_file()}\n{e}')
    except input_.LocationNotFound as e:
        show_warning(f'Location from input file not found. Please fix it and try again\n'
                     f'{input_.get_last_input_file()}\n{e}')
    return []


def _get_temp_file_name() -> str:
    return input_.get_watched_file_name() + '.tmp'


def _get_duration_str(ms: int, override_ms: int) -> str:
    hours, remainder = divmod(ms, 1000 * 60 * 60)
    minutes, remainder = divmod(remainder, 1000 * 60)
    seconds, remainder = divmod(remainder, 1000)
    hours_override = override_ms / (1000 * 60 * 60)
    hours_len = 2 if hours_override <= 0 else max(ceil(log10(hours_override)), 2)
    hours_fmt = '{' + ':0{}d'.format(hours_len) + '}'
    return (hours_fmt + ':{:02d}:{:02d}').format(hours, minutes, seconds)



#    Interleave Playlist
#    Copyright (C) 2021-2022 Thomas Sweeney
#    This file is part of Interleave Playlist.
#    Interleave Playlist is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    Interleave Playlist is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import itertools
import math
import os
import subprocess
import threading
import typing
from typing import Optional, Callable, Any

import natsort
from PySide6.QtCore import Slot, QEvent, Qt, Signal, QThread, QDeadlineTimer, SignalInstance
from PySide6.QtGui import QFont, QColor, QBrush, QFontDatabase, QCloseEvent
from PySide6.QtWidgets import QVBoxLayout, QListWidget, QWidget, QAbstractItemView, QHBoxLayout, \
    QPushButton, QMessageBox, QFileDialog, QLabel, QGridLayout, QProgressBar, QRadioButton, \
    QGroupBox, QCheckBox, QLineEdit, QLayout
from natsort import natsorted
from pymediainfo import MediaInfo

from interleave_playlist.core.playlist import PlaylistEntry
from interleave_playlist.interface import open_with_default_application, _create_playlist, \
    _get_duration_str
from interleave_playlist.interface.PlaylistWindowItem import PlaylistWindowItem
from interleave_playlist.interface.SearchBarThread import SearchBarThread, \
    SearchBarThreadAlreadyDeadException
from interleave_playlist.persistence import input_
from interleave_playlist.persistence import settings
from interleave_playlist.persistence.input_ import set_last_input_file
from interleave_playlist.persistence.watched import add_watched, remove_watched

_LIGHT_MODE_WATCHED_COLOR = QBrush(QColor.fromRgb(255, 121, 121))
_DARK_MODE_WATCHED_COLOR = QBrush(QColor.fromRgb(77, 12, 12))
_TOTAL_SHOWS_TEXT = 'Total Shows:    {}'
_SELECTED_SHOWS_TEXT = 'Selected Shows: {}'
_TOTAL_RUNTIME = 'Total Runtime:    {}'
_SELECTED_RUNTIME = 'Selected Runtime: {}'


class RuntimeCalculationThread(QThread):
    value_updated = typing.cast(SignalInstance, Signal(int))
    completed = typing.cast(SignalInstance, Signal(dict, int))
    error = typing.cast(SignalInstance, Signal(BaseException))

    def __init__(self,
                 playlist: list[PlaylistEntry],
                 duration_cache: Optional[dict[str, int]] = None):
        super(RuntimeCalculationThread, self).__init__()
        if duration_cache is None:
            duration_cache = {}
        self.running: bool = False
        self.stop: bool = False
        self.playlist: list[PlaylistEntry] = playlist.copy()
        self.duration_cache: dict[str, int] = ({}
                                               if duration_cache is None else
                                               duration_cache.copy())
        self.pending_playlist: Optional[list[PlaylistEntry]] = None

    def __del__(self) -> None:
        self.wait()

    def set_pending_playlist(self, playlist: list[PlaylistEntry]) -> None:
        self.pending_playlist = playlist.copy()

    def run(self) -> None:
        self.running = True
        try:
            self._run()
        except BaseException as e:
            self.error.emit(e)
        finally:
            self.running = False

    def _run(self) -> None:
        while True:
            total_duration = 0
            for i, elem in enumerate(item.filename for item in self.playlist):
                if self.stop:
                    return
                if elem not in self.duration_cache:
                    duration: int = 0
                    if os.path.isfile(elem):
                        media_info = MediaInfo.parse(elem)
                        if len(media_info.video_tracks) > 0 \
                                and media_info.video_tracks[0].duration:
                            duration = int(float(media_info.video_tracks[0].duration))
                        elif len(media_info.audio_tracks) > 0 \
                                and media_info.audio_tracks[0].duration:
                            duration = int(float(media_info.audio_tracks[0].duration))
                        else:
                            print(f'Warning: {elem} has no duration info for video or audio')
                            duration = 0
                    self.duration_cache[elem] = duration
                    self.value_updated.emit(i+1)
                total_duration += self.duration_cache[elem]
            if not self.pending_playlist:
                break
            self.playlist = self.pending_playlist
            self.pending_playlist = None
        self.completed.emit(self.duration_cache, total_duration)


class PlaylistWindow(QWidget):
    def __init__(self) -> None:
        super().__init__()
        self._warned_about_mediainfo_missing = False
        self.selection_dependent_buttons: list[QPushButton] = []
        self._row_color1, self._row_color2 = self._get_standard_row_colors()
        self.duration_cache: dict[str, int] = {}
        self.durations_loaded = False
        counter = itertools.count()
        self.sort: Callable[[Any], Any] = lambda x: next(counter)

        self.search_bar_thread = None
        label_font = QFontDatabase.systemFont(QFontDatabase.FixedFont)
        label_font.setPointSize(int(settings.get_font_size() * 1.25))
        self.total_shows_label = QLabel()
        self.total_shows_label.setFont(label_font)

        self.total_selected_label = QLabel()
        self.total_selected_label.setFont(label_font)

        self.playlist: list[PlaylistEntry] = []
        self.item_list: QListWidget = self._create_item_list()

        self.total_runtime_label = QLabel(_TOTAL_RUNTIME.format('...'))
        self.total_runtime_label.setFont(label_font)
        self.total_runtime_progress = QProgressBar()
        self.total_runtime_progress.setMaximum(len(self.playlist))
        self.total_runtime_progress.hide()

        self.selected_runtime_label = QLabel(_SELECTED_RUNTIME.format('...'))
        self.selected_runtime_label.setFont(label_font)
        self.item_list.selectAll()

        self.runtime_thread = None
        self.total_duration: int = 0

        search_label = QLabel("Search ")
        self.search_bar = QLineEdit()
        self.search_bar.textEdited.connect(self.search_bar_text_edited)
        self.search_bar.editingFinished.connect(self.search_bar_editing_finished)
        self.search_bar.setToolTip("Ctrl+F")
        search_layout = QHBoxLayout()
        search_layout.addWidget(search_label)
        search_layout.addWidget(self.search_bar)

        self.interleave_radio = QRadioButton("Interleave")
        self.interleave_radio.setToolTip('Ctrl+Shift+I')
        self.interleave_radio.toggle()
        self.interleave_radio.toggled.connect(self.interleave_sort)
        self.alphabetical_radio = QRadioButton("Alphabetical")
        self.alphabetical_radio.setToolTip('Ctrl+Shift+A')
        self.alphabetical_radio.toggled.connect(self.alphabetical_sort)
        self.last_modified_radio = QRadioButton("Last Modified")
        self.last_modified_radio.setToolTip('Ctrl+Shift+M')
        self.last_modified_radio.toggled.connect(self.last_modified_sort)
        self.reversed_checkbox = QCheckBox("Reversed")
        self.reversed_checkbox.toggled.connect(self.reverse_sort)
        self.reversed_checkbox.setToolTip('Ctrl+Shift+R')

        total_layout = QVBoxLayout()
        stats_group = QGroupBox("Stats")
        total_layout.addWidget(self.total_shows_label)
        total_layout.addWidget(self.total_selected_label)
        runtime_layout = QVBoxLayout()
        runtime_layout.addWidget(self.total_runtime_label)
        runtime_layout.addWidget(self.selected_runtime_label)
        stats_layout = QHBoxLayout()
        stats_layout.addLayout(total_layout)
        stats_layout.addLayout(runtime_layout)
        stats_group.setLayout(stats_layout)
        label_layout = QHBoxLayout()
        label_layout.addWidget(stats_group, 4)

        radio_group = QGroupBox("Search && Sort")
        radio_layout = QGridLayout()
        radio_layout.addWidget(self.interleave_radio,    0, 0)
        radio_layout.addWidget(self.alphabetical_radio,  1, 0)
        radio_layout.addWidget(self.last_modified_radio, 0, 1)
        radio_layout.addWidget(self.reversed_checkbox,   1, 1)
        radio_layout.addLayout(search_layout,            2, 0, 1, 2)
        radio_group.setLayout(radio_layout)

        label_layout.addWidget(radio_group, 2)

        list_layout = QVBoxLayout()
        list_layout.addWidget(self.item_list)

        layout = QVBoxLayout(self)
        layout.addWidget(self.total_runtime_progress)
        layout.addLayout(label_layout)
        layout.addLayout(self._create_button_layout())
        layout.addLayout(list_layout)

        self._refresh()
        self.item_list.selectAll()

    def _create_item_list(self) -> QListWidget:
        item_list = QListWidget()
        item_list.setSelectionMode(QAbstractItemView.ExtendedSelection)
        item_list.setAlternatingRowColors(True)
        font = QFont()
        font.setPointSize(settings.get_font_size())
        item_list.setFont(font)
        item_list.doubleClicked.connect(self.play)
        item_list.itemSelectionChanged.connect(self.selection_change)
        item_list.installEventFilter(self)
        return item_list

    def _create_button_layout(self) -> QLayout:
        button_layout = QHBoxLayout()
        buttons = [
            ('Play', self.play, 'Enter', True),
            ('Mark Watched', self.mark_watched, 'Ctrl-W', True),
            ('Unmark Watched', self.unmark_watched, 'Ctrl-U', True),
            ('Drop Shows', self.drop_groups, 'Ctrl-Shift-D', True),
            ('Refresh', self.refresh, 'F5', False),
            ('Open Input File', self.open_input, 'Ctrl-O', False),
            ('Open Watched File', self.open_watched_file, 'Ctrl-Shift-O', False),
        ]
        for name, func, tooltip, selection_dependent in buttons:
            button = QPushButton(name)
            button.clicked.connect(func)
            if tooltip is not None:
                button.setToolTip(tooltip)
            if selection_dependent:
                self.selection_dependent_buttons.append(button)
            button_layout.addWidget(button)
        return button_layout

    @typing.no_type_check
    def eventFilter(self, widget: QWidget, event: QEvent) -> bool:
        if event.type() == QEvent.KeyPress:
            switch: dict[int, Callable] = {
                Qt.NoModifier + Qt.Key_Return: self._play,
                Qt.NoModifier + Qt.Key_Enter: self._play,
                Qt.ControlModifier + Qt.Key_W: self.mark_watched,
                Qt.ControlModifier + Qt.Key_U: self.unmark_watched,
                Qt.NoModifier + Qt.Key_F5: self.refresh,
                Qt.ControlModifier + Qt.ShiftModifier + Qt.Key_D: self.drop_groups,
                Qt.ControlModifier + Qt.Key_O: self.open_input,
                Qt.ControlModifier + Qt.ShiftModifier + Qt.Key_O: self.open_watched_file,
                Qt.ControlModifier + Qt.ShiftModifier + Qt.Key_I: self.interleave_radio.toggle,
                Qt.ControlModifier + Qt.ShiftModifier + Qt.Key_A: self.alphabetical_radio.toggle,
                Qt.ControlModifier + Qt.ShiftModifier + Qt.Key_M: self.last_modified_radio.toggle,
                Qt.ControlModifier + Qt.ShiftModifier + Qt.Key_R: self.reversed_checkbox.toggle,
                Qt.ControlModifier + Qt.Key_F: self._focus_search_bar
            }
            if event.keyCombination().toCombined() in switch:
                switch[event.keyCombination().toCombined()]()
                return True
        return False

    @Slot()
    def play(self) -> None:
        self._play()
        self.item_list.setFocus()

    def _play(self) -> None:
        def _impl() -> None:
            files = [i.getValue().filename
                     for i
                     in typing.cast(list[PlaylistWindowItem], self.item_list.selectedItems())]
            subprocess.run([settings.get_play_command()] + files)
        if self.playlist is not None:
            thread = threading.Thread(target=_impl)
            thread.start()

    @Slot()
    def mark_watched(self) -> None:
        selected_values: list[PlaylistEntry] = [
            i.getValue()
            for i
            in typing.cast(list[PlaylistWindowItem], self.item_list.selectedItems())
        ]
        add_watched(selected_values)
        for item in self.item_list.selectedItems():
            item.setBackground(self._get_watched_color())
        self.item_list.setFocus()

    @Slot()
    def unmark_watched(self) -> None:
        selected_values: list[PlaylistEntry] = [
            i.getValue()
            for i
            in typing.cast(list[PlaylistWindowItem], self.item_list.selectedItems())
        ]
        remove_watched(selected_values)
        for i, item in enumerate(self.item_list.selectedItems()):
            color = self._row_color1 if i % 2 == 0 else self._row_color2
            item.setBackground(color)
        self.item_list.setFocus()

    @Slot()
    def drop_groups(self) -> None:
        selected_entries: list[PlaylistEntry] = [
            i.getValue()
            for i
            in typing.cast(list[PlaylistWindowItem], self.item_list.selectedItems())
        ]
        if len(selected_entries) == 0:
            return
        groups_str = set()
        for entry in selected_entries:
            groups_str.add(entry.group.name)
        msg_box = QMessageBox()
        msg_box.setWindowTitle('Drop Groups')
        msg_box.setText("You are about to drop the following groups. "
                        "Do you wish to continue?\n    {}"
                        .format('\n    '.join(natsorted(groups_str))))
        msg_box.setIcon(QMessageBox.Icon.Question)
        msg_box.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
        reply = msg_box.exec()
        if reply == QMessageBox.Ok:
            input_.drop_groups(selected_entries)
        self.item_list.setFocus()

    @Slot()
    def refresh(self) -> None:
        self._refresh()

    def _refresh(self, *, use_cache: bool = False) -> None:
        self.item_list.clear()
        self.playlist = _create_playlist(self.search_bar.text(), use_cache)
        if self.playlist is not None:
            for item in sorted(self.playlist,
                               key=self.sort,
                               reverse=self.reversed_checkbox.isChecked()):
                self.item_list.addItem(
                    PlaylistWindowItem(value=item)
                )
        self.total_shows_label.setText(_TOTAL_SHOWS_TEXT.format(len(self.playlist)))
        self.durations_loaded = False
        if self.item_list.count() > 0:
            self.item_list.setCurrentItem(self.item_list.item(0))
        self._run_calculate_total_runtime_thread()
        self._refresh_buttons()
        self.item_list.setFocus()

    @Slot()
    def open_input(self) -> None:
        try:
            file_name: str = QFileDialog.getOpenFileName(
                self, 'Open yaml', os.path.expanduser('~'), 'yaml files (*.yml *.yaml)')[0]
            if file_name.strip() == '':
                return
            try:
                set_last_input_file(file_name)
            except input_.InvalidInputFile:
                msg_box = QMessageBox()
                msg_box.setWindowTitle('Error')
                msg_box.setText(f'Error: Attempted to open invalid input yaml file\n\n{file_name}')
                msg_box.setIcon(QMessageBox.Icon.Warning)
                msg_box.exec()
                return
            self._stop_runtime_thread()
            self._refresh()
        finally:
            self.item_list.setFocus()

    @Slot()
    def open_watched_file(self) -> None:
        def _impl() -> None:
            open_with_default_application(input_.get_watched_file_name())
        thread = threading.Thread(target=_impl)
        thread.start()
        self.item_list.setFocus()

    @Slot()
    def selection_change(self) -> None:
        self._selection_change(len(self.item_list.selectedItems()))

    @Slot()
    def update_total_runtime_progress_bar(self, value: int) -> None:
        if self.runtime_thread is None:
            return
        self.total_runtime_progress.setMaximum(len(self.runtime_thread.playlist))
        self.total_runtime_progress.setValue(value)

    @Slot()
    def total_runtime_thread_completed(self, duration_cache: dict[str, int],
                                       total_duration: int) -> None:
        self.duration_cache = duration_cache
        self.total_duration = total_duration
        self.total_runtime_label.setText(
            _TOTAL_RUNTIME.format(_get_duration_str(total_duration, total_duration)))
        self.durations_loaded = True
        self._get_selected_runtime()
        self.total_runtime_progress.hide()

    @Slot()
    def total_runtime_thread_error(self, exception: BaseException) -> None:
        raise exception

    @Slot()
    def search_bar_thread_error(self, exception: BaseException) -> None:
        raise exception

    @Slot()
    def interleave_sort(self, checked: bool) -> None:
        if not checked:
            return
        counter = itertools.count()
        self.sort = lambda x: next(counter)
        self._refresh(use_cache=True)
        self.item_list.setFocus()

    @Slot()
    def alphabetical_sort(self, checked: bool) -> None:
        if not checked:
            return
        self.sort = natsort.natsort_keygen(
            lambda i: os.path.basename(i.filename))  # type: ignore[no-any-return]
        self._refresh(use_cache=True)
        self.item_list.setFocus()

    @Slot()
    def last_modified_sort(self, checked: bool) -> None:
        if not checked:
            return
        self.sort = lambda x: os.path.getmtime(x.filename)
        self._refresh(use_cache=True)
        self.item_list.setFocus()

    @Slot()
    def reverse_sort(self, checked: bool) -> None:
        self._refresh(use_cache=True)
        self.item_list.setFocus()

    @Slot()
    def search_bar_text_edited(self, text: str) -> None:
        if self.search_bar_thread is None:
            self._init_search_bar_thread(text)
            return
        try:
            self.search_bar_thread.search_value = text
            return
        except SearchBarThreadAlreadyDeadException:
            self._init_search_bar_thread(text)

    @Slot()
    def search_bar_editing_finished(self) -> None:
        self._refresh(use_cache=True)
        # we don't want focus back here. we're finished

    @Slot()
    def search_bar_thread_completed(self, text: str) -> None:
        self._refresh(use_cache=True)
        self.search_bar.setFocus()

    def _init_search_bar_thread(self, text: str) -> None:
        self.search_bar_thread = SearchBarThread(text, 500)
        self.search_bar_thread.error.connect(self.search_bar_thread_error)
        self.search_bar_thread.completed.connect(self.search_bar_thread_completed)
        self.search_bar_thread.start()

    def _focus_search_bar(self) -> None:
        self.search_bar.setFocus()
        self.search_bar.selectAll()

    def _selection_change(self, num_selected: int) -> None:
        self.total_selected_label.setText(
            _SELECTED_SHOWS_TEXT.format(
                str(num_selected).rjust(math.ceil(
                    math.log10(len(self.playlist))
                    if len(self.playlist) > 0 else
                    1
                ))))
        self._get_selected_runtime()
        for btn in self.selection_dependent_buttons:
            btn.setEnabled(num_selected > 0)

    def _get_selected_runtime(self) -> None:
        if not self.durations_loaded:
            return
        self.selected_runtime_label.setText(_SELECTED_RUNTIME.format('...'))
        duration = sum([self.duration_cache[i.getValue().filename]
                        for i
                        in typing.cast(list[PlaylistWindowItem], self.item_list.selectedItems())])
        self.selected_runtime_label.setText(
            _SELECTED_RUNTIME.format(_get_duration_str(duration, self.total_duration)))

    @staticmethod
    def _get_standard_row_colors() -> tuple[QBrush, QBrush]:
        # wow this is silly
        _hidden_list = QListWidget()
        _hidden_list.setAlternatingRowColors(True)
        _hidden_list.addItems(["", ""])
        color1 = _hidden_list.item(0).background()
        color2 = _hidden_list.item(1).background()
        return color1, color2

    @staticmethod
    def _get_watched_color() -> QBrush:
        return (_LIGHT_MODE_WATCHED_COLOR
                if not settings.get_dark_mode() else
                _DARK_MODE_WATCHED_COLOR)

    def closeEvent(self, event: QCloseEvent) -> None:
        self._stop_runtime_thread()

    def _run_calculate_total_runtime_thread(self) -> None:
        self.selected_runtime_label.setText(_SELECTED_RUNTIME.format('...'))
        self.durations_loaded = False
        if not MediaInfo.can_parse():
            if not self._warned_about_mediainfo_missing:
                self._warned_about_mediainfo_missing = True
                msg_box = QMessageBox()
                msg_box.setWindowTitle('Warning')
                msg_box.setText("libmediainfo not found. Will be unable to calculate "
                                "runtimes unless installed")
                msg_box.setIcon(QMessageBox.Icon.Warning)
                msg_box.exec()
            return
        if self.runtime_thread is not None and not self.runtime_thread.isFinished():
            self.runtime_thread.set_pending_playlist(self.playlist)
            return
        self.total_runtime_progress.setValue(0)
        self.total_runtime_progress.show()
        self.total_runtime_label.setText(_TOTAL_RUNTIME.format('...'))
        self.runtime_thread = RuntimeCalculationThread(self.playlist, self.duration_cache)
        self.runtime_thread.value_updated.connect(self.update_total_runtime_progress_bar)
        self.runtime_thread.completed.connect(self.total_runtime_thread_completed)
        self.runtime_thread.error.connect(self.total_runtime_thread_error)
        self.runtime_thread.start()

    def _stop_runtime_thread(self) -> None:
        if self.runtime_thread is None:
            return

        self.runtime_thread.stop = True
        self.runtime_thread.wait(QDeadlineTimer(10 * 1000))
        self.runtime_thread = None
        self.durations_loaded = False
        self.duration_cache = {}

    def _refresh_buttons(self) -> None:
        for button in self.selection_dependent_buttons:
            button.setEnabled(len(self.item_list.selectedItems()) > 0)



#    Interleave Playlist
#    Copyright (C) 2021-2022 Thomas Sweeney
#    This file is part of Interleave Playlist.
#    Interleave Playlist is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    Interleave Playlist is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
from os import path
from typing import Any

from PySide6.QtWidgets import QListWidgetItem

from interleave_playlist.core.playlist import PlaylistEntry

_USER_TYPE: int = 1001


class PlaylistWindowItem(QListWidgetItem):

    def __init__(self, *args: list[Any], value: PlaylistEntry, **kwargs: dict[Any, Any]):
        super().__init__(*args, type=_USER_TYPE, **kwargs)  # type: ignore
        self.value: PlaylistEntry = value
        self.setText(path.basename(value.filename))

    def setValue(self, value: PlaylistEntry) -> None:
        self.value = value

    def getValue(self) -> PlaylistEntry:
        return self.value



#    Interleave Playlist
#    Copyright (C) 2021-2022 Thomas Sweeney
#    This file is part of Interleave Playlist.
#    Interleave Playlist is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    Interleave Playlist is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path
import sys
from traceback import format_exception
from types import TracebackType
from typing import Type, Optional, Sequence

from PySide6.QtWidgets import QApplication, QMessageBox

from interleave_playlist import SCRIPT_LOC
from interleave_playlist.interface.PlaylistWindow import PlaylistWindow
from interleave_playlist.persistence.settings import get_dark_mode


class PlaylistApplication(QApplication):
    def __init__(self, arr: Sequence[str]):
        super().__init__(arr)
        playlist_window = PlaylistWindow()
        playlist_window.setWindowTitle('Interleave Playlist')
        playlist_window.resize(800, 600)
        playlist_window.show()
        if get_dark_mode():
            with open(os.path.join(
                    SCRIPT_LOC,
                    'interface',
                    'style',
                    'dark.qss'
            ), 'r') as f:
                _style = f.read()
                self.setStyleSheet(_style)
        sys.exit(self.exec())


def excepthook(cls: Type[BaseException], exception: BaseException,
               traceback: Optional[TracebackType]) -> None:
    err = "".join(format_exception(cls, exception, traceback))
    msg_box = QMessageBox()
    msg_box.setWindowTitle("Error")
    msg_box.setText(f'Encountered an unhandled exception. This is a bug.\n'
                    f'Please file a bug report so that this can be fixed\n\n{err}')
    msg_box.setIcon(QMessageBox.Icon.Critical)
    msg_box.exec()


sys.excepthook = excepthook



#    Interleave Playlist
#    Copyright (C) 2021-2022 Thomas Sweeney
#    This file is part of Interleave Playlist.
#    Interleave Playlist is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    Interleave Playlist is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import pathlib
import sys
from dataclasses import dataclass, field
from datetime import timedelta, datetime
from typing import Optional

import appdirs
from crontab import CronTab

from interleave_playlist.persistence.settings import _create_settings_file

_STATE_FILE = pathlib.Path(os.path.join(appdirs.user_config_dir(),
                                        'interleave_playlist',
                                        'state.json'))


class Timed:
    def __init__(self, start: datetime, cron: CronTab, first: Optional[int] = None,
                 amount: Optional[int] = None, start_at_cron: Optional[bool] = False):
        self.start = (start.astimezone()
                      if start.tzinfo is None else
                      start)
        self.cron = cron
        self.first = (first if first is not None and first > 0 else 1) - 1
        self.amount = amount if amount is not None and amount > 0 else 1
        self.start_at_cron = start_at_cron

    def __repr__(self) -> str:
        return str(self.__dict__)

    def get_current(self) -> int:
        now: datetime = datetime.now().astimezone()
        if self.start > now:
            return -1
        initial: int = self.first + (self.amount if not self.start_at_cron else 0)
        if self.start_at_cron and self.cron.test(self.start):
            first_diff = timedelta(seconds=0)
        else:
            first_diff = timedelta(seconds=self.cron.next(self.start, default_utc=False))
        first_cron_amount: int = self.amount if self.start + first_diff <= now else 0
        if first_cron_amount == 0 and self.start_at_cron:
            return -1
        diff = timedelta(seconds=self.cron.next(self.start + first_diff, default_utc=False))
        cron_amount = self.amount * max((now - (self.start + first_diff)) // diff, 0)
        return initial + first_cron_amount + cron_amount - 1


@dataclass(unsafe_hash=True)
class Group:
    name: str = field(hash=True)
    location_name: str = field(default="", hash=True)
    priority: int = field(default=sys.maxsize, hash=False, compare=False)
    whitelist: list[str] = field(default_factory=list, hash=False, compare=False)
    blacklist: list[str] = field(default_factory=list, hash=False, compare=False)
    timed: Optional[Timed] = field(default=None, hash=False, compare=False)

    def __post_init__(self) -> None:
        if self.priority is None:
            self.priority = sys.maxsize
        if self.whitelist is None:
            self.whitelist = []
        if self.blacklist is None:
            self.blacklist = []


@dataclass(unsafe_hash=True)
class Location:
    name: str = field(hash=True)
    default_group: Group = field(hash=False)
    additional: list[str] = field(default_factory=list, hash=False)
    regex: Optional[str] = field(default=None, hash=False)
    groups: list[Group] = field(default_factory=list, hash=False)

    def __post_init__(self) -> None:
        if self.groups is None:
            self.groups = []
        if self.additional is None:
            self.additional = []


def _create_state_file() -> None:
    if not os.path.exists(_STATE_FILE):
        with open(_STATE_FILE, 'w') as f:
            f.write('{"last-input-file": "See config/input.yml.example"}')


def create_needed_files() -> None:
    _create_settings_file()
    _create_state_file()



#    Interleave Playlist
#    Copyright (C) 2021-2022 Thomas Sweeney
#    This file is part of Interleave Playlist.
#    Interleave Playlist is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    Interleave Playlist is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
import typing
from typing import Any

import appdirs
from ruamel.yaml import YAML

_SETTINGS_FILE = pathlib.Path(os.path.join(appdirs.user_config_dir(),
                                           'interleave_playlist',
                                           'settings.yml'))
_CACHED_FILE: dict[str, Any] = {}


def get_font_size() -> int:
    return typing.cast(int, _get_settings('font-size'))


def get_play_command() -> str:
    return typing.cast(str, _get_settings('play-command'))


def get_dark_mode() -> bool:
    return typing.cast(bool, _get_settings('dark-mode'))


def get_max_watched_remembered() -> int:
    return typing.cast(int, _get_settings('max-watched-remembered'))


def get_exclude_directories() -> bool:
    return typing.cast(bool, _get_settings('exclude-directories'))


def _get_settings(option: str) -> Any:
    global _CACHED_FILE
    if not _CACHED_FILE:
        with open(_SETTINGS_FILE, 'r') as f:
            yaml = YAML()
            yaml.preserve_quotes = True  # type: ignore
            _CACHED_FILE = yaml.load(f)
            default_settings = _get_default_settings()
        for key in default_settings.keys():
            if key not in _CACHED_FILE:
                _CACHED_FILE[key] = default_settings[key]
        with open(_SETTINGS_FILE, 'w') as f:
            yaml.dump(_CACHED_FILE, f)

    return _CACHED_FILE[option]


def _get_default_settings() -> dict[str, Any]:
    return {
        'font-size': 12,
        'play-command': 'mpv',
        'dark-mode': False,
        'max-watched-remembered': 100,
        'exclude-directories': True
    }


def _create_settings_file() -> None:
    if not os.path.exists(_SETTINGS_FILE):
        if not os.path.exists(_SETTINGS_FILE.parent):
            os.mkdir(_SETTINGS_FILE.parent)
        with open(_SETTINGS_FILE, 'w') as f:
            yaml = YAML()
            yaml.dump(_get_default_settings(), f)



#    Interleave Playlist
#    Copyright (C) 2021-2022 Thomas Sweeney
#    This file is part of Interleave Playlist.
#    Interleave Playlist is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    Interleave Playlist is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import re
from copy import copy
from itertools import groupby
from os import path
from re import Pattern
from typing import Any

from natsort import natsorted, ns

from interleave_playlist.core import PlaylistEntry
from interleave_playlist.core.interleave import interleave_all
from interleave_playlist.persistence import Group, Location, Timed
from interleave_playlist.persistence import settings

FilePathsByGroup = dict[Group, list[str]]
FileGroup = tuple[str, str]
PlaylistEntriesByGroup = dict[Group, list[PlaylistEntry]]
_FILE_CACHE: dict[str, list[str]] = {}


def get_playlist(locations: list[Location],
                 watched_list: list[FileGroup],
                 search_filter: str = "",
                 use_cache: bool = False) -> list[PlaylistEntry]:
    location_groups: PlaylistEntriesByGroup = {}
    for loc in locations:
        paths: list[str] = _get_paths_from_location(loc, use_cache)
        location_groups.update(_group_items_by_regex(loc, paths))

    def _key(i: tuple[Group, list[PlaylistEntry]]) -> int: return i[0].priority
    entries_by_priority: dict[int, PlaylistEntriesByGroup] = {
        k: dict(v) for k, v in groupby(sorted(location_groups.items(), key=_key), _key)
    }
    data: list[PlaylistEntry] = []
    for ep in entries_by_priority.values():
        data.extend(_get_playlist(ep, watched_list, search_filter))
    return data


def _get_playlist(entries_by_group: PlaylistEntriesByGroup,
                  watched_list: list[FileGroup],
                  search_filter: str = "") -> list[PlaylistEntry]:
    filtered_entries: list[list[PlaylistEntry]] = []
    watched_names = [i[0].upper() for i in watched_list]
    for group, entries in entries_by_group.items():
        # Ordering is important! Filter out invalid considerations first
        group_entries = [entry for entry in filter(
            lambda i: (_matches_whitelist(path.basename(i.filename), group.whitelist)
                       and not _matches_blacklist(path.basename(i.filename), group.blacklist)
                       and (not settings.get_exclude_directories() or os.path.isfile(i.filename))),
            entries)]
        # Now that invalid considerations are gone, we can slice by timing considerations
        # or else invalid considerations will be part of the result, then removed anyway
        group_entries = _timed_slice(group.timed, group_entries) if group.timed else group_entries
        # Now that invalid and timed considerations are gone, we can finally remove things
        # that we've already seen and match by the search filter
        group_entries = [
            entry for entry in filter(
                lambda i: path.basename(i.filename).upper() not in watched_names
                and search_filter.upper() in path.basename(i.filename).upper(),
                group_entries
            )
        ]
        if group_entries:
            filtered_entries.append(group_entries)
    if not filtered_entries:
        return []
    # Need to do some convoluted nonsense to remove alphabetical biasing in the playlist.
    # Need to make sure the playlist is ordered by least recently watched in a way
    # that doesn't interfere with the quality of the interleaving.
    masked: list[list[tuple[int, int]]] = _mask_data(filtered_entries)
    masked_playlist: list[tuple[int, int]] = interleave_all(masked)
    sorted_group: list[list[PlaylistEntry]] = \
        _sort_data_by_least_recently_watched(filtered_entries, watched_list)
    return _unmask_playlist(masked_playlist, sorted_group)


def _get_paths_from_location(loc: Location, use_cache: bool) -> list[str]:
    if use_cache and loc.name in _FILE_CACHE:
        return _FILE_CACHE[loc.name]
    path_parts = [(loc.name, i) for i in os.listdir(loc.name)]
    for a in loc.additional:
        path_parts += [(a, i) for i in os.listdir(a)]
    paths = list(map(lambda i: path.join(i[0], i[1]),
                     natsorted(path_parts, key=lambda i: i[1], alg=ns.IGNORECASE)))
    _FILE_CACHE[loc.name] = paths
    return paths


def _group_items_by_regex(loc: Location, paths: list[str]) -> PlaylistEntriesByGroup:
    regex_str: str = loc.regex if loc.regex is not None else ''
    regex: Pattern = re.compile(regex_str)
    grouped_items: PlaylistEntriesByGroup = {}
    group_dict = {group.name.upper(): group for group in loc.groups}

    for p in paths:
        match = regex.match(path.basename(p))
        if not match:
            continue
        match_dict = match.groupdict()
        if 'group' in match_dict:
            group_name = path.basename(match.group('group')).strip()
            group = _get_from_dict_key_superset(group_name.upper(), group_dict)
            if group is None:
                group = copy(loc.default_group)
                group_dict[group_name.upper()] = group
            # Overwrite name so we always use the full name
            group.name = group_name
        else:
            group = loc.default_group
        group_members = grouped_items.setdefault(group, list())
        group_members.append(PlaylistEntry(p, loc, group))
    return grouped_items


def _timed_slice(timed: Timed, loc_list: list[PlaylistEntry]) -> list[PlaylistEntry]:
    cur_release = min(timed.get_current(), len(loc_list))
    if cur_release < 0:
        return []
    return loc_list[timed.first: cur_release + 1]


def _matches_whitelist(s: str, whitelist: list[str]) -> bool:
    if not whitelist:
        return True
    for white in whitelist:
        if white.upper() in s.upper():
            return True
    return False


def _matches_blacklist(s: str, blacklist: list[str]) -> bool:
    if not blacklist:
        return False
    for black in blacklist:
        if black and black.upper() in s.upper():
            return True
    return False


def _get_from_dict_key_superset(super_key: str, d: dict[str, Any]) -> Any:
    for k, v in d.items():
        if k in super_key:
            return v


def _get_watched_groups_lru(watched_list: list[FileGroup]) -> list[str]:
    groups: list[str] = []
    for i in reversed(watched_list):
        if not i[1] in groups:
            groups.append(i[1])
    groups.reverse()
    return groups


def _mask_data(data: list[list[PlaylistEntry]]) -> list[list[tuple[int, int]]]:
    masked: list[list[tuple[int, int]]] = []
    group_idx: dict[int, int] = {}
    for group in data:
        masked_group: list[tuple[int, int]] = []
        masked.append(masked_group)
        idx = group_idx.setdefault(len(group), 0)
        for _ in group:
            masked_group.append((idx, len(group)))
        group_idx[len(group)] = idx + 1
    return masked


def _sort_data_by_least_recently_watched(
        data: list[list[PlaylistEntry]],
        watched_list: list[FileGroup]) -> list[list[PlaylistEntry]]:
    lru_groups: list[str] = _get_watched_groups_lru(watched_list)
    ideal_group_sorting: list[list[PlaylistEntry]] = []
    for i, d in enumerate(data):
        if d[0].group.name not in lru_groups:
            ideal_group_sorting.append(d)
    for group in lru_groups:
        for i, d in enumerate(data):
            if d[0].group.name == group:
                ideal_group_sorting.append(d)
    return ideal_group_sorting


def _unmask_playlist(masked_playlist: list[tuple[int, int]], data: list[list[PlaylistEntry]]) \
        -> list[PlaylistEntry]:
    playlist: list[PlaylistEntry] = []
    data_size_dict: dict[int, list[list[PlaylistEntry]]] = {}
    group_len_group_idx_dict: dict[int, dict[int, int]] = {}
    group_file_idx_dict: dict[tuple[int, int], int] = {}
    for group in data:
        sized_data: list[list[PlaylistEntry]] = data_size_dict.setdefault(len(group), [])
        sized_data.append(group)
    for masked_group in masked_playlist:
        group_idx, group_len = masked_group
        group_idx_dict = group_len_group_idx_dict.setdefault(group_len, {})
        mapped_group_idx = group_idx_dict.setdefault(group_idx, len(group_idx_dict))
        group_file_idx = group_file_idx_dict.setdefault(masked_group, 0)
        group_file_idx_dict[masked_group] = group_file_idx + 1

        playlist.append(data_size_dict[group_len][mapped_group_idx][group_file_idx])
    return playlist



#    Interleave Playlist
#    Copyright (C) 2022 Thomas Sweeney
#    This file is part of Interleave Playlist.
#    Interleave Playlist is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    Interleave Playlist is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import typing
from threading import Lock
from time import time_ns

from PySide6.QtCore import QThread, Signal, SignalInstance


class SearchBarThreadAlreadyDeadException(BaseException):
    pass


class SearchBarThread(QThread):
    completed = typing.cast(SignalInstance, Signal(str))
    error = typing.cast(SignalInstance, Signal(BaseException))

    def __init__(self, search_value: str, delay_ms: int):
        super(SearchBarThread, self).__init__()
        if delay_ms < 0:
            raise ValueError("delay_ms can not be negative")
        if search_value is None:
            raise ValueError("value must not be null")
        self._search_value = search_value
        self._delay_ns = delay_ms * 1000 * 1000
        self._wait_until = time_ns() + self._delay_ns
        self._lock = Lock()
        self._has_run = False
        self._running = False

    def __del__(self) -> None:
        self.wait()

    @property
    def search_value(self) -> str:
        return self._search_value

    @search_value.setter
    def search_value(self, search_value: str) -> None:
        self._lock.acquire()
        if not self._running and self._has_run:
            raise SearchBarThreadAlreadyDeadException()
        try:
            self._search_value = search_value
            self._wait_until = time_ns() + self._delay_ns
        finally:
            self._lock.release()

    def run(self) -> None:
        try:
            self._run()
        except BaseException as e:
            self.error.emit(e)
            self._lock.release()

    def _run(self) -> None:
        # _running and _has_run must be set in this order!!!
        self._running = True
        self._has_run = True
        while True:
            self._lock.acquire()
            now = time_ns()
            wait_until_copy = self._wait_until
            if self._wait_until <= now:
                break
            self._lock.release()
            self.msleep((wait_until_copy - now) // 1000 // 1000)
        self.completed.emit(self._search_value)
        self._running = False
        self._lock.release()
